import math
import random
import pgzrun
import pickle

#-------- Game entities

class Seeker(Actor):
  def __init__(self):
    global boss_alive
    #moves toward player
    Actor.__init__(self, 'bluemissle1', (WIDTH + 50, 200))
    self.timer = 0
    self.health = 1
    boss_alive = True

  def update(self, enemy_container, bullet_container, player_bullet_container, player):
    global boss_alive
    #animate
    if self.image == 'bluemissle1':
      self.image = 'bluemissle2'
    if self.image == 'bluemissle2':
      self.image = 'bluemissle3'
    if self.image == 'bluemissle3':
      self.image = 'bluemissle4'
    if self.image == 'bluemissle4':
      self.image = 'bluemissle1'

    #Move left on screen move toward player
    if player.y > self.y:
      self.y += 0.7
    elif player.y < self.y:
      self.y -= 0.7

    self.x -= 0.7

    #Update timer for firing
    self.timer += 1

    #Remove self if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 100:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)

  def draw(self):
    Actor.draw(self)

class Bonus(Actor):
  def __init__(self, type, xpos, ypos):
    if type == 'random':
      type_list = ['credit', 'shield']
      self.type = random.choice(type_list)
    else:
      self.type = type

    Actor.__init__(self, self.type, (xpos, ypos))
    self.xdir = -3
    self.ydir = -3

  def update(self, bonus_container):
    if self.y > HEIGHT and self.ydir > 0:
      self.ydir = -3
    elif self.y < 0 and self.ydir < 0:
      self.ydir = 3

    self.y += self.ydir
    self.x += self.xdir

    if self.x < -80:
      bonus_container.remove(self)

  def draw(self):
    Actor.draw(self)

class Bullet(Actor):
  def __init__(self, posx, posy, angle, speed, is_player):
    #Sets the image to a blue ball if it is a player bullet 
    #Or to an orange ball if it is an enemy bullet
    if is_player is True:
      Actor.__init__(self, 'player_bullet', (posx, posy))
    else:
      Actor.__init__(self, 'enemy_bullet', (posx, posy))

    #Calculate direction bullet will travel from angle provided
    self.dirx = speed * math.cos(math.radians(angle))
    self.diry = speed * math.sin(math.radians(angle))

  def update(self, bullet_container):
    #Move bullet towards desired direction
    self.x += self.dirx
    self.y += self.diry

    #Remove bullet from its container if it goes offscreen
    if self.x > 850 or self.x < -50 or self.y > 650 or self.y < -50:
      bullet_container.remove(self)

  def draw(self):
    #Draw bullet to screen at current position
    Actor.draw(self)

class Coin(Actor):
  def __init__(self, xpos, ypos):
    #Starts the coin at the provided location
    Actor.__init__(self, 'credit', (xpos, ypos))

    #Send coin in a random direction from where it spawns
    angle = 2 * math.pi * random.random()
    self.dirx = math.cos(angle) * random.randint(1, 2)
    self.diry = math.sin(angle) * random.randint(1, 2)

    #Initialize the timer used for magnet timing
    self.timer = 0

  def update(self, coin_container, player):
    #Simulate gravity for a little less than 1 second
    #Then magnet the coin towards the player if they are in range
    self.timer += 1
    if self.timer < 50:
      if self.diry > 0:
        self.y += self.diry * 1.3
    else:
      if Actor.distance_to(self, player) < 400:
        angle = math.atan2(player.y - self.y, player.x - self.x)
        self.dirx = math.cos(angle + 0.2) * 6
        self.diry = math.sin(angle + 0.2) * 6
      else:
        self.dirx = 0
        self.diry = 3

    #Remove coin from its container if it falls completely off the screen
    if self.y > 700:
      coin_container.remove(self)

    #Move coin towards desired direction
    self.x += self.dirx
    self.y += self.diry

class Feather(Actor):
  def __init__(self):
    #Select a left facing or right facing feather at random
    image = random.choice(['feather', 'feather2'])
    
    #Spawn the feather at a randomized position at the top of the screen
    xpos = random.randrange(0, WIDTH)
    ypos = random.randrange(-200, -100)

    #Create an actor using the previously created variables
    Actor.__init__(self, image, (xpos, ypos))

    #Initialize timer for calculating feather position
    self.timer = 0

    #Randomize the direction that the feather is initialy traveling in
    self.dir = random.choice(['left', 'right'])

  def update(self, feather_container):
    #Swap direction that feather is traveling every 20 frames
    self.timer += 1
    if self.timer == 20:
      if self.dir == 'left':
        self.dir = 'right'
      else:
        self.dir = 'left'

      self.timer = 0

    #Move feather in the direction that it is traveling
    if self.dir == 'left':
      self.x -= 2
    else:
      self.x += 2

    #Apply gravity to make the feather fall down the screen
    self.y += 2

    #Remove feather from its container when it falls off of the screen
    if self.y > 700:
      feather_container.remove(self)

  def draw(self):
    #Draw the feather to the screen at its current position
    Actor.draw(self)

class Enemy_1(Actor):
  def __init__(self):
    #Start the enemy at a random position just to the right off screen
    Actor.__init__(self, 'ball', (WIDTH + 50, random.randrange(40, 480)))

    #Initialize timer for firing rate
    self.timer = 0

    #Start the enemy with 6 health points
    self.health = 6

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Move left on screen
    self.x -= 2

    #Remove enemy from its container if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Fire bullet towards player every 20 frames
    self.timer += 1
    if self.timer == 20 and self.x > 300:
      #Calculate angle between enemy and player
      angle = math.degrees(math.atan2(player.y - self.y, player.x - self.x))

      #Fire one bullet in that towards the player
      bullet_container.append(Bullet(self.x, self.y, angle, 5.5, False))
      bullet_container.append(Bullet(self.x, self.y, angle, 3.5, False))

      #Reset fire rate timer
      self.timer = 0

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 60:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)

        #Reduce the enemy health
        self.health -= 1

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 30
      chain += 10
      chain_timer = 120

      #Create an explosion of 15 coins
      for i in range(0, 15):
        coin_container.append(Coin(self.x, self.y))

  def draw(self):
    #Draw the enemy to the screen at its current position
    Actor.draw(self)

class Enemy_2(Actor):
  def __init__(self):
    #Start the enemy at a random position just to the right off screen
    Actor.__init__(self, 'ball_sm', (WIDTH + 50, random.randrange(50, 480)))

    #Initialize timer for firing rate
    self.timer = 0

    #Start the enemy with 3 health points
    self.health = 3

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Move left on screen while oscillating up and down
    self.x -= 4
    self.y += 3 * math.sin(self.x / 60)

    #Remove enemy from container if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Fire bullet towards player every 30 frames
    self.timer += 1
    if self.timer == 30 and self.x > 300:
      #Calculate angle between enemy and player
      angle = math.degrees(math.atan2(player.y - self.y, player.x - self.x))

      #Fire three bullets in a spread in the direction of the player
      bullet_container.append(Bullet(self.x, self.y, angle + 5, 5, False))
      bullet_container.append(Bullet(self.x, self.y, angle - 5, 5, False))
      bullet_container.append(Bullet(self.x, self.y, angle, 5, False))

      #Reset fire rate timer
      self.timer = 0

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 40:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 10
      chain += 5
      chain_timer = 120
      
      #Create an explosion of 10 coins
      for i in range(0, 10):
        coin_container.append(Coin(self.x, self.y))

  def draw(self):
    #Draw the enemy to the screen at its current position
    Actor.draw(self)

class Enemy_3(Actor):
  def __init__(self):
    #Start the enemy at a random position just to the right off screen
    Actor.__init__(self, 'ball', (WIDTH + 50, random.randrange(40, 480)))

    #Initialize timer for firing rate
    self.timer = 0

    #Start the enemy with 4 health points
    self.health = 4

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Move left on screen
    self.x -= 2

    #Update timer for firing
    self.timer += 1

    #Remove self if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Fire bullet to the left every 15 frames
    if self.timer % 15 == 0 and self.x > 300:
      #Increase the speed of bullets fired slowly over time
      speed = self.timer / 6
      bullet_container.append(Bullet(self.x, self.y, 180, speed, False))
      bullet_container.append(Bullet(self.x, self.y, 15, speed / 4, False))
      bullet_container.append(Bullet(self.x, self.y, -15, speed / 4, False))

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 60:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 30
      chain += 10
      chain_timer = 120

      #Create an explosion of 18 coins
      for i in range(0, 18):
        coin_container.append(Coin(self.x, self.y))

  def draw(self):
    #Draw the enemy to the screen at its current position
    Actor.draw(self)

class Enemy_4(Actor):
  def __init__(self):
    #Start the enemy at a random position just to the right off screen
    Actor.__init__(self, 'ball', (WIDTH + 50, random.randrange(40, 480)))

    #Initialize timer for firing rate
    self.timer = 0

    #Start the enemy with 4 health points
    self.health = 3

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Move left on screen
    self.x -= 6

    #Update timer for firing
    self.timer += 2

    #Remove self if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Fire bullet to the left every 4 frames
    if self.timer % 4 == 0 and self.x > 500:
      #Fire bullets
      bullet_container.append(Bullet(self.x, self.y, self.timer * 5, 5, False))
      bullet_container.append(Bullet(self.x, self.y, 360 - self.timer * 5, 5, False))

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 60:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 30
      chain += 10
      chain_timer = 120

      #Create an explosion of 18 coins
      for i in range(0, 18):
        coin_container.append(Coin(self.x, self.y))

  def draw(self):
    #Draw the enemy to the screen at its current position
    Actor.draw(self)

class Enemy_5(Actor):
  def __init__(self):
    #Start the enemy at a random position just to the right off screen
    Actor.__init__(self, 'ball', (WIDTH + 50, random.randrange(40, 480)))

    #Initialize timer for firing rate
    self.timer = 0

    #Start the enemy with 6 health points
    self.health = 2

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Move left on screen
    self.x -= 4

    #Remove enemy from its container if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Fire bullet towards player every 20 frames
    self.timer += 1
    if self.timer == 40 and self.x > 300:
      #Fire bullets
      for i in range(0, 360, 40):
        bullet_container.append(Bullet(self.x, self.y, i, 5, False))

      #Reset fire rate timer
      self.timer = 0

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 60:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)

        #Reduce the enemy health
        self.health -= 1

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 30
      chain += 10
      chain_timer = 120

      #Create an explosion of 15 coins
      for i in range(0, 15):
        coin_container.append(Coin(self.x, self.y))

  def draw(self):
    #Draw the enemy to the screen at its current position
    Actor.draw(self)

class Boss1(Actor):
  def __init__(self):
    #Spawn the boss off to the right of the screen
    Actor.__init__(self, 'ball', (WIDTH + 50, 200))

    #Initialize timer for firing rate
    self.timer = 0

    #Start the boss with 100 health points
    self.max_health = 100
    self.health = 100

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Move left on screen but stop a little ways in
    if self.x > 600:
      self.x -= 0.7

    #Oscillate up and down
    self.y += math.sin(self.timer / 10) * 5

    #Remove self if off screen
    if self.x < -80:
      enemy_container.remove(self)

    #Fire bullets towards player every 15 frames
    self.timer += 1
    if self.timer % 20 == 0:
      #Calculate angle between enemy and player
      angle = math.degrees(math.atan2(player.y - self.y, player.x - self.x))
      #Fire five bullets in a spread in the direction of the player
      bullet_container.append(Bullet(self.x, self.y, angle + 28, 4, False))
      bullet_container.append(Bullet(self.x, self.y, angle + 14, 4, False))
      bullet_container.append(Bullet(self.x, self.y, angle - 14, 4, False))
      bullet_container.append(Bullet(self.x, self.y, angle - 28, 4, False))
      bullet_container.append(Bullet(self.x, self.y, angle, 4,False))
      #Fire a spread of bullets
      for i in range(0, 360, 60):
        bullet_container.append(Bullet(self.x, self.y, i, 3, False))

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 100:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1
        chain += 1
        chain_timer = 120

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 100

      #Create an explosion of 35 coins
      for i in range(0, 35):
        coin_container.append(Coin(self.x, self.y))

      #Turn all bullets on screen into coins
      for bullet in bullet_container:
        coin_container.append(Coin(bullet.x, bullet.y))

      bullet_container.clear()

      #Spawn a shield collectable
      bonus_container.append(Bonus('shield', self.x, self.y))

  def draw(self):
    #Draw the boss to the screen at its current position
    Actor.draw(self)

class Boss2(Actor):
  def __init__(self):
    #Spawn the boss in the bottom right off screen
    Actor.__init__(self, 'ball', (WIDTH + 50, HEIGHT + 50))

    #Initialize timer for firing rate and movement
    self.timer = 0

    #Start the boss with 100 health points
    self.max_health = 150
    self.health = 150

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Update timer for movement and fire rate
    self.timer += 1

    #Handle boss movement
    if self.timer < 60:
      self.x -= 2
      self.y -= 3
    else:
      self.x += math.cos(self.timer / 20) * 3
      self.y += math.sin(self.timer / 20) * 5

    #Fire bullets towards player
    if self.timer % 15 == 0:
      #Calculate angle between enemy and player
      angle = math.degrees(math.atan2(player.y - self.y, player.x - self.x))

      #Fire many bullets
      for i in range(0, 360, 40):
        bullet_container.append(Bullet(self.x, self.y, angle + i, 3, False))
        bullet_container.append(Bullet(self.x, self.y, angle + i, 2, False))

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 100:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1
        chain += 1
        chain_timer = 120

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 150

      #Create an explosion of 40 coins
      for i in range(0, 40):
        coin_container.append(Coin(self.x, self.y))

      #Turn all bullets on screen into coins
      for bullet in bullet_container:
        coin_container.append(Coin(bullet.x, bullet.y))

      bullet_container.clear()

      #Spawn a shield collectable
      bonus_container.append(Bonus('shield', self.x, self.y))

  def draw(self):
    #Draw the boss to the screen at its current position
    Actor.draw(self)

class Boss3(Actor):
  def __init__(self):
    #Spawn the boss in the bottom right off screen
    Actor.__init__(self, 'ball', (WIDTH + 50, HEIGHT + 50))

    #Initialize timer for firing rate and movement
    self.timer = 0

    #Start the boss with 100 health points
    self.max_health = 150
    self.health = 150
    
    self.angle = 180

  def update(self, enemy_container, bullet_container, player_bullet_container,
             player, coin_container, bonus_container):
    global score
    global chain
    global chain_timer

    #Update timer for movement and fire rate
    self.timer += 1

    #Handle boss movement
    if self.timer < 60:
      self.x -= 3
      self.y -= 6

    #Fire bullets towards player
    if self.timer % 5 == 0 and self.timer > 60:
      #Calculate angle between enemy and player
      angle = math.degrees(math.atan2(player.y - self.y, player.x - self.x))

      for i in range(0, 360, 48):
        bullet_container.append(Bullet(self.x, self.y, i + self.timer * 10, 4, False))
        bullet_container.append(Bullet(self.x, self.y, -i - self.timer * 10, 4, False))

    #Handle collision against player bullets
    for bullet in player_bullet_container:
      if Actor.distance_to(self, bullet) < 100:
        #Remove the player bullet that collided with the enemy
        player_bullet_container.remove(bullet)
        #Reduce the enemy health
        self.health -= 1
        chain += 1
        chain_timer = 120

    #Handle death if health is 0 or less
    if self.health <= 0:
      enemy_container.remove(self)
      score += 150

      #Create an explosion of 40 coins
      for i in range(0, 40):
        coin_container.append(Coin(self.x, self.y))

      #Turn all bullets on screen into coins
      for bullet in bullet_container:
        coin_container.append(Coin(bullet.x, bullet.y))

      bullet_container.clear()

      #Spawn a shield collectable
      bonus_container.append(Bonus('shield', self.x, self.y))

  def draw(self):
    #Draw the boss to the screen at its current position
    Actor.draw(self)

class Player(Actor):
  def __init__(self):
    #Start the player towards the left of the screen
    Actor.__init__(self, 'tbird_1', (300, 200))

    #Timer for fire rate
    self.timer = 0

    #Timer for handling animation
    self.timer2 = 0

    #Player variables
    self.speed = 8
    self.shield = 2

    #Variables for handling invincibility
    self.invincible = False
    self.invincible_timer = 0

  def update(self, bullet_container, player_bullet_container, bonus_container, coin_container):
    global score
    global chain
    global chain_timer

    #Update timer for player invincibility
    self.invincible_timer -= 1

    #Handle player movement
    dirx = 0
    diry = 0
    if keyboard.left:
      dirx -= 1
    if keyboard.right:
      dirx += 1
    if keyboard.up:
      #Play animation when play is flying upwards
      self.timer2 += 1
      if self.image == 'tbird_1' and self.timer2 == 4:
        self.image = 'tbird_2'
        self.timer2 = 0
      elif self.image == 'tbird_2' and self.timer2 == 4:
        self.image = 'tbird_3'
        self.timer2 = 0
      elif self.image == 'tbird_3' and self.timer2 == 4:
        self.image = 'tbird_4'
        self.timer2 = 0
      elif self.image == 'tbird_4' and self.timer2 == 4:
        self.image = 'tbird_1'
        self.timer2 = 0
      diry -= 1
    if keyboard.down:
      diry += 1

    #Normalize player movement
    if dirx != 0 and diry != 0:
      dirx = 0.707 if dirx > 0 else -0.707
      diry = 0.707 if diry > 0 else -0.707

    #Move player towards direction
    self.x += dirx * self.speed
    self.y += diry * self.speed

    #Handle player firing when Z is down
    self.timer += 1
    if keyboard.z and self.timer > 10:
      player_bullet_container.append(Bullet(self.x, self.y, 0, 25, True))
      player_bullet_container.append(Bullet(self.x, self.y, -10, 25, True))
      player_bullet_container.append(Bullet(self.x, self.y, 10, 25, True))
      #Reset fire rate timer
      self.timer = 0

    #Make sure the player does not go off screen
    if self.y < 0:
      self.y = 0
    elif self.y > 600:
      self.y = 600
    if self.x < 0:
      self.x = 0
    elif self.x > 800:
      self.x = 800

    #Reset wings
    if self.image == 'tbird_2':
      self.image = 'tbird_3'
    elif self.image == 'tbird_3':
      self.image = 'tbird_4'
    elif self.image == 'tbird_4':
      self.image = 'tbird_1'

    #Handle bonuses
    for bonus in bonus_container:
      if Actor.distance_to(self, bonus) < 30:
        if bonus.type == 'shield':
          self.shield += 1
        bonus_container.remove(bonus)

    #Handle coins
    for coin in coin_container:
      if Actor.distance_to(self, coin) < 30:
        score += chain
        coin_container.remove(coin)

    #Check collision against bullets
    for bullet in bullet_container:
      if Actor.distance_to(self, bullet) < 10:
        if self.invincible_timer <= 0:
          if self.shield > 0:
            self.shield -= 1
            self.invincible_timer = 180
            chain = 0
            chain_timer = 0
          else:
            scene_stack.pop()
            scene_stack.append(Scene_Game_Over())
            return
        
        bullet_container.remove(bullet)

  def draw(self):
    #Draw player at current position with flickering effect if invincible
    if self.invincible_timer <= 0 or self.invincible_timer % 4 <= 1:
      Actor.draw(self)

    #Draw a little circle over the player to convey where the collider is
    screen.draw.filled_circle((self.x, self.y), 7, "red")

#-------- Game scenes

class Scene_Game:
  def __init__(self):
    global score
    global chain
    global chain_timer

    self.timer = 0
    self.timer_rate = 51
    score = 0
    chain = 0
    chain_timer = 0

    self.layer5 = Actor('layer5', topleft=(0, 0))
    self.layers = []
    self.layers.append(Actor('layer4', topleft=(0, 0)))
    self.layers.append(Actor('layer3', topleft=(0, 0)))
    self.layers.append(Actor('layer2', topleft=(0, 0)))
    self.layers.append(Actor('layer1', topleft=(0, 0)))

    #Game entities
    self.enemies = []
    self.bullets = []
    self.player_bullets = []
    self.player = Player()
    self.boss = None
    self.bonus = []
    self.coins = []

  def update(self):
    global scene_stack
    global highscore
    global score
    global chain
    global chain_timer

    #Update chain
    chain_timer -= 1

    if chain_timer <= 0:
      chain = 0

    #Update background parallax scrolling effect
    self.layers[0].x -= 2
    self.layers[1].x -= 4
    self.layers[2].x -= 8
    self.layers[3].x -= 16

    for layer in self.layers:
      if layer.x <= 0:
        layer.x = 1066

    #Update high score if current score is higher
    if score > highscore:
      highscore = score

    #Pause the game if escape is pressed
    if keyboard.escape:
      scene_stack.append(Scene_Pause())

    #Add an enemy to the game every interval
    if self.timer == self.timer_rate:
      num = random.randint(1, 65)
      if num > 62:
        if num == 65:
          self.boss = Boss1()
        elif num == 64:
          self.boss = Boss2()
        elif num == 63:
          self.boss = Boss3()

        if self.timer_rate > 24:
          self.timer_rate -= 3
        self.enemies.append(self.boss)
      elif num > 50:
        self.enemies.append(Enemy_1())
      elif num > 40:
        self.enemies.append(Enemy_2())
      elif num > 30:
        self.enemies.append(Enemy_3())
      elif num > 20:
        self.enemies.append(Enemy_4())  
      else:
        self.enemies.append(Enemy_5())
      self.timer = 0

    #Increase timer for adding enemies if there is no boss alive
    if self.boss not in self.enemies:
      self.timer += 1

    #Update all entities
    self.player.update(self.bullets, self.player_bullets, self.bonus, 
                       self.coins)

    for bullet in self.bullets:
      bullet.update(self.bullets)

    for bullet in self.player_bullets:
      bullet.update(self.player_bullets)

    for abonus in self.bonus:
      abonus.update(self.bonus)

    for coin in self.coins:
      coin.update(self.coins, self.player)

    for enemy in self.enemies:
      enemy.update(self.enemies, self.bullets, self.player_bullets, 
                   self.player, self.coins, self.bonus)

  def draw(self):
    #Draw background
    self.layer5.draw()
    for layer in self.layers:
      layer.draw()

    #Draw all game entities to the screen
    for bullet in self.player_bullets:
      bullet.draw()

    for enemy in self.enemies:
      enemy.draw()

    for coin in self.coins:
      coin.draw()

    for abonus in self.bonus:
      abonus.draw()

    self.player.draw()

    for bullet in self.bullets:
      bullet.draw()

    #Display player statistics at top of screen
    screen.draw.text("Score: " + str(score),
                     midleft=(20, 20),
                     owidth=1.5,
                     fontsize=30)
    screen.draw.text("Shield: " + str(self.player.shield),
                     midleft=(20, 40),
                     owidth=1.5,
                     fontsize=30)
    screen.draw.text("HighScore: " + str(highscore),
                     center=(WIDTH / 2, 20),
                     owidth=1.5,
                     fontsize=30)

    #Display boss health bar if boss is alive
    if self.boss in self.enemies:
      width = 500 / self.boss.max_health * self.boss.health
      screen.draw.filled_rect(Rect(150, 50, 500, 30), "grey")
      screen.draw.filled_rect(Rect(150, 50, width, 30),"red")
      screen.draw.rect(Rect(150, 50, 500, 30), "black")

    #Display chain information
    screen.draw.text("Chain: " + str(chain),
                     midleft=(20, 520),
                     owidth=1.5,
                     fontsize=30)
    width = 200 / 120 * chain_timer
    screen.draw.filled_rect(Rect(20, 550, 200, 30), "grey")
    screen.draw.filled_rect(Rect(20, 550, width, 30),"green")
    screen.draw.rect(Rect(20, 550, 200, 30), "black")

class Scene_Menu:
  def __init__(self):
    #Images that will be displayed behind title
    self.background = Actor('layer5', topleft=(0, 0))
    self.tbird = Actor('start_tbird', (170, 200))

  def update(self):
    global scene_stack

    #Enter game when z is pressed
    if keyboard.z:
      scene_stack.pop()
      scene_stack.append(Scene_Game())

  def draw(self):
    #Draw background art
    self.background.draw()
    self.tbird.draw()

    #Draw title to the screen
    screen.draw.text("T-Bird Fighter!",
                     center=(WIDTH / 2, 380),
                     fontsize=100,
                     color="red",
                     shadow=(1, 1),
                     scolor="#202020")
    
    #Display control instructions
    screen.draw.text("Controls:", (540, 480))
    screen.draw.text("Arrow keys to move", (540, 500))
    screen.draw.text("Press Z to shoot", (540, 520))
    screen.draw.text("Press Escape to pause", (540, 540))

    screen.draw.text("Press Z to start game! ", (60, 470), fontsize=60)
    screen.draw.text( "HighScore: " + str(highscore), (60, 520), fontsize=40)

class Scene_Pause:
  def __init__(self):
    pass

  def update(self):
    global scene_stack

    #Unpause game when return is pressed
    if keyboard.RETURN:
      scene_stack.pop()

  def draw(self):
    #Draw previous scene. Should be Game_Scene
    scene_stack[-2].draw()

    #Draw text over screen
    screen.draw.text("Paused", center=(WIDTH / 2, 250), fontsize=100, owidth=2)
    screen.draw.text("Press Enter to resume game",
                     center=(WIDTH / 2, 300),
                     fontsize=40,
                     owidth=2)

class Scene_Game_Over:
  def __init__(self):
    #Feather container
    self.feathers = []
    
    #Background art
    self.background = Actor('game_over_back', topleft=(0, 0))

    #Initialize timer for feather spawning
    self.timer = 0

    #Save score to file if it is the current highest score
    if score == highscore:
      with open('score.dat', 'wb') as file:
        pickle.dump(highscore, file)

  def update(self):
    global scene_stack
    
    #Create a new feather every 22 frames
    self.timer += 1
    if self.timer == 22:
      self.feathers.append(Feather())
      self.timer = 0

    #Enter game when z is pressed
    if keyboard.RETURN:
      scene_stack.pop()
      scene_stack.append(Scene_Menu())

    #Update all feathers
    for feather in self.feathers:
      feather.update(self.feathers)

  def draw(self):
    #Draw background art
    self.background.draw()

    #Draw feathers to screen
    for feather in self.feathers:
      feather.draw()
    
    #Draw game over text
    screen.draw.text("Game Over!",
                     center=(WIDTH / 2, HEIGHT / 2),
                     color="red",
                     shadow=(2, 1),
                     fontsize=60)
    screen.draw.text("Press Enter to return to main menu",
                     center=(WIDTH / 2, 350),
                     owidth=1.5,
                     fontsize=40)
    screen.draw.text("Score: " + str(score),
                     center=(WIDTH / 2, 400),
                     owidth=1.5,
                     fontsize=40)

#-------- Initialization

score = 0
highscore = 0
chain = 0
chain_timer = 0

#Window dimensions
WIDTH = 800
HEIGHT = 600

#Stack of all scenes
scene_stack = []

#Start the game at the start menu
scene_stack.append(Scene_Menu())

#Read highscore from file
try:
  with open('score.dat', 'rb') as file:
    highscore = pickle.load(file)
except:
  highscore = 0

#-------- Pygame zero game loop function declarations

def update():
  #Update the scene at the top of the scene stack
  scene_stack[-1].update()

def draw():
  screen.clear()
  #Draw the scene at the top of the scene stack
  scene_stack[-1].draw()

pgzrun.go()
